import {CashuMint, CashuWallet, getEncodedToken} from "@cashu/cashu-ts";
import {logger} from "./utils";
import * as nip98 from "nostr-tools/nip98";
import {Elysia, t} from "elysia";
import {swagger} from "@elysiajs/swagger";
import {serverTiming} from "@elysiajs/server-timing";
import {PrismaClient} from "@prisma/client";
import {TokenInfoWithMailSubscriptionDuration} from "@arx/utils/cashu.ts";
import {npubToPubKeyString, pubKeyStringToNpub} from "@arx/utils/nostr.ts";
import cors from "@elysiajs/cors";

const npubType = t.String({
  pattern: `^npub1[023456789acdefghjklmnpqrstuvwxyz]{58}$`,
  error: 'Invalid npub format'
});

const cashuTokenType = t.String({
  pattern: '^cashu[A-Za-z0-9+-_]*={0,3}$',
  error: 'Invalid Cashu token format'
})

export class HttpServer {
  constructor(private db: PrismaClient, port: number) {
    new Elysia()
      .use(swagger({
        documentation: {
          info: {
            title: 'npub.email Documentation',
            version: '0.0.1'
          }
        }
      }))
      .use(serverTiming())
      .use(cors())
      .get('/', 'nostr.email server')
      .get('/subscription/:npub', this.getSubscriptionForNpub, {
        params: t.Object({
          npub: npubType
        })
      })
      .get('/aliases/:npub', this.getAliasesForNpub, {
        params: t.Object({
          npub: npubType,
        }),
      })
      .get('/alias/:alias', this.getNpubForAlias, {
        params: t.Object({
          alias: t.String(),
        }),
      })
      .post('/addAlias', this.addAlias, {
        body: t.Object({
          alias: t.String()
        })
      })
      .post('/addTime/:npub', this.addTimeToNpub, {
        params: t.Object({
          npub: npubType,
        }),
        body: t.Object({
          tokenString: cashuTokenType
        })
      })
      .listen(port)
    logger.info(`HTTP Server running on port ${port}`);
  }

  getSubscriptionForNpub = async ({params: {npub}}: {
    params: {
      npub: string
    }
  }) => {
    const user = await this.db.user.findFirst({
      where: {
        npub
      },
      include: {
        aliases: true
      }
    });
    if (!user) return {
      subscribed: false
    };
    return {
      subscribed: true,
      subscribedUntil: user.subscriptionDuration == null ? Infinity : Math.floor(user.lastPayment.getTime() / 1000) + user.subscriptionDuration
    };
  }

  getNpubForAlias = async ({params: {alias}}: {
    params: {
      alias: string
    }
  }) => {
    const user = await this.db.user.findFirst({
      where: {
        aliases: {
          some: {
            alias
          }
        }
      }
    });
    if (!user) return new Response('Not found', {
      status: 404
    });
    return user.npub;
  }

  getAliasesForNpub = async ({params: {npub}, headers}: {
    params: {
      npub: string
    },
    headers: Record<string, string | undefined>
  }) => {
    const unpacked = await this.getUnpackedAuthHeader(headers, `/aliases/${npub}`);
    const npubAsPubkey = npubToPubKeyString(npub);
    if (unpacked.pubkey !== npubAsPubkey)
      return new Response('Unauthorized', {
        status: 401
      })
    const user = await this.db.user.findFirst({
      where: {
        npub
      },
      include: {
        aliases: true
      }
    });
    if (!user) return new Response('Not found', {
      status: 404
    });
    return user.aliases.map(alias => alias.alias);
  }

  addAlias = async ({body: {alias}, headers}: {
    body: {
      alias: string
    },
    headers: Record<string, string | undefined>
  }) => {
    const unpacked = await this.getUnpackedAuthHeader(headers, '/addAlias');
    const unpackedKeyToNpub = pubKeyStringToNpub(unpacked.pubkey);
    const userInDb = await this.db.user.findFirst({
      where: {
        npub: unpackedKeyToNpub
      }
    });
    if (!userInDb) return new Response('Unauthorized', {
      status: 401
    });

    const stillHasSubscription = userInDb.subscriptionDuration === null || Math.floor(userInDb.lastPayment.getTime() / 1000) + userInDb.subscriptionDuration > Date.now() / 1000;
    if (!stillHasSubscription) return new Response('User has no subscription', {
      status: 400
    });
    const aliasInDb = await this.db.alias.findFirst({
      where: {
        alias
      }
    });
    if (aliasInDb) return new Response('Alias already exists', {
      status: 400
    });
    return this.db.user.update({
      where: {
        npub: unpackedKeyToNpub
      },
      data: {
        aliases: {
          create: {
            alias
          }
        }
      }
    });
  }

  addTimeToNpub = async ({params: {npub}, body: {tokenString}}: {
    params: {
      npub: string
    },
    body: {
      tokenString: string
    }
  }) => {
    const userInDb = await this.db.user.findFirst({
      where: {
        npub
      }
    });

    if (userInDb && (userInDb.subscriptionDuration === null || userInDb.subscriptionDuration === -1))
      return new Response('User has unlimited subscription', {
        status: 400
      })

    const tokenInfo = new TokenInfoWithMailSubscriptionDuration(tokenString);
    const mint = new CashuMint(tokenInfo.mint);
    const wallet = new CashuWallet(mint);
    const newToken = await wallet.receive(tokenString);
    const encodedToken = getEncodedToken({
      token: [{
        mint: tokenInfo.mint,
        proofs: newToken
      }]
    });
    logger.info(`New cashu token: ${encodedToken}`);
    if (userInDb) {
      let timeRemaining = Math.max(0, Math.floor((+new Date(userInDb.lastPayment.getTime() + userInDb.subscriptionDuration! * 1000) - +new Date()) / 1000));
      timeRemaining += tokenInfo.duration;
      await this.db.user.update({
        where: {
          npub
        },
        data: {
          lastPayment: new Date(),
          subscriptionDuration: timeRemaining
        }
      });
      return {
        newTimeRemaining: timeRemaining
      }
    }
    await this.db.user.create({
      data: {
        npub,
        registeredAt: new Date(),
        lastPayment: new Date(),
        subscriptionDuration: tokenInfo.duration
      }
    });
    return {
      newTimeRemaining: tokenInfo.duration
    }
  }

  private getUnpackedAuthHeader = async (headers: Record<string, string | undefined>, url: string) => {
    if (!headers.authorization)
      throw new Error('Unauthorized');
    const authHeader = headers.authorization.split(' ')[1];
    const validate = await nip98.validateToken(authHeader, `${process.env.PUBLIC_API_BASE_URL!}${url}`, "POST");
    if (!validate)
      throw new Error('Unauthorized');
    return await nip98.unpackEventFromToken(authHeader);
  }
}
